#include "pch.hpp"
#include "SwrAddrDef.h"

#define POLL_INTERVAL 500

static enum SFMSTATE {
	SFM_FINDWINDOW,
	SFM_WATCHSWR,
};

static HANDLE s_ThProc;
static HWND   s_ThWnd;

static HWND s_callbackWnd;
static int  s_callbackMsg;

static HANDLE s_userThread;
static HANDLE s_thread;
static DWORD s_threadId;

static int s_paramOld[SWRPARAM_MAX];
static char s_paramStr[260];

static void APIENTRY SwrAPCCallback(ULONG_PTR)
{
}

static void SwrCallback(short Msg, short param1, int param2)
{
	if(s_callbackWnd) {
		::PostMessage(s_callbackWnd, s_callbackMsg, MAKELONG(Msg, param1), param2);
	}
}


static SFMSTATE SwrStateFindWindow()
{
	s_ThWnd = ::FindWindow(SWR_WINDOW_CLASS, SWR_WINDOW_TEXT);
	if(s_ThWnd != NULL) {
		DWORD dwThId;
		::GetWindowThreadProcessId(s_ThWnd, &dwThId);
		s_ThProc = ::OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwThId);
		SwrCallback(SWRMSG_STATECHANGE, SWRSTATE_WATCH, 0);
		::ZeroMemory(s_paramOld, sizeof s_paramOld);
		return SFM_WATCHSWR;
	} else {
		return SFM_FINDWINDOW;
	}
}

static SFMSTATE SwrStateWatchSwr()
{
	DWORD ret = ::WaitForSingleObject(s_ThProc, 0);
	if(ret != WAIT_TIMEOUT) {
		::CloseHandle(s_ThProc);
		s_ThProc = NULL;
		SwrCallback(SWRMSG_STATECHANGE, SWRSTATE_NOTFOUND, 0);
		return SFM_FINDWINDOW;
	}

	int param;
	for(int i = 0; i < SWRPARAM_MAX; ++i) {
		if((param = SwrGetParam(i)) != -1) {
			if(param != s_paramOld[i])
				SwrCallback(SWRMSG_PARAMCHANGE, i, param);
			s_paramOld[i] = param;
		} else s_paramOld[i] = 0;
	}
	return SFM_WATCHSWR;
}

static DWORD WINAPI SwrAddrWorkThread(LPVOID)
{
	MSG msg;
	SFMSTATE sfmState = SFM_FINDWINDOW;

	while(!::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) || msg.message != WM_QUIT) {
		switch(sfmState) {
		case SFM_FINDWINDOW: sfmState = SwrStateFindWindow(); break;
		case SFM_WATCHSWR:   sfmState = SwrStateWatchSwr();   break;
		};
		::SleepEx(POLL_INTERVAL, TRUE);
	}
	::ExitThread(0);
	return 0;
}

int SwrAddrInit(HWND callbackWnd, int callbackMsg)
{
	s_callbackWnd = callbackWnd;
	s_callbackMsg = callbackMsg;
	s_thread = ::CreateThread(NULL, 0, SwrAddrWorkThread, NULL, 0, &s_threadId);
	s_userThread = ::GetCurrentThread();
	return s_thread != NULL;
}

int SwrAddrFinish()
{
	if(s_thread) {
		::PostThreadMessage(s_threadId, WM_QUIT, 0, 0);
		::QueueUserAPC(SwrAPCCallback, s_thread, 0);
		::WaitForSingleObject(s_thread, INFINITE);
		::CloseHandle(s_thread);
	}
	return 0;
}

int SwrGetParam(int param)
{
	DWORD dwReadSize;
	DWORD dwordData;
	BYTE byteData;

	switch(param) {
	case SWRPARAM_SCENE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_SCENEID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_COMMMODE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_COMMMODE, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_BATTLEMODE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_BTLMODEOFS), &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_LCHARID:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_LCHARID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_RCHARID:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_RCHARID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_LWINCOUNT:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_LCHAROFS), &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_WINCNTOFS), &byteData, 1, &dwReadSize)) {
			return byteData;
		}
		break;
	case SWRPARAM_RWINCOUNT:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_RCHAROFS), &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_WINCNTOFS), &byteData, 1, &dwReadSize)) {
			return byteData;
		}
		break;
	case SWRPARAM_LPROFNAME:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
			::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_LPROFOFS), s_paramStr, SWR_ADDR_PROFSIZE, &dwReadSize)) {
			return (int)s_paramStr;
		}
		break;
	case SWRPARAM_RPROFNAME:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
			::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_RPROFOFS), s_paramStr, SWR_ADDR_PROFSIZE, &dwReadSize)) {
			return (int)s_paramStr;
		}
		break;
	case SWRPARAM_TOADDR:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
			::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_TOADDROFS), &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	}
	return -1;
}
