#include "pch.hpp"
#include "mainwnd.hpp"
#include "chalcopyrite.hpp"
#include "scoreline.hpp"
#include "scorelinedlg.hpp"
#include "Formatter.hpp"
#include "TextFileWriter.hpp"
#include "SwrAddrDef.h"
#include "resource.h"

#define MINIMAL_USE_PROCESSHEAPSTRING
#include "MinimalPath.hpp"

#define UM_NOTIFYICON (WM_APP + 1)
#define UM_SWRCALLBACK (WM_APP + 10)


static UINT s_WM_TASKBAR_CREATED;
static NOTIFYICONDATA s_nid;
static HMENU s_hPopupMenu;
static HWND  s_hScoreLineDlg;
static BOOL  s_disableSwrCallback;
static BOOL  s_isHage;				/* ÎВ */
static BOOL  s_tcgLimit;
static int   s_tcgCount;
static int   s_mode;

static void Swr_OutputLog(LPCSTR lprof, int lid, int lwin, LPCSTR rprof, int rid, int rwin)
{
	Minimal::ProcessHeapStringT<char> baseName;
	baseName = ScoreLine_GetPath();
	::PathStripPath(baseName);
	::PathRemoveExtension(baseName);
	baseName.Repair();

	Minimal::ProcessHeapPathA logPath = g_appPath;
	logPath /= baseName;
	if(
	!PathIsDirectory(logPath) &&
	!CreateDirectory(logPath, NULL)) return;

	SYSTEMTIME locTime;
	GetLocalTime(&locTime);

	logPath /=
		Formatter("%04d%02d%02d.csv",
			locTime.wYear, locTime.wMonth, locTime.wDay);
	TextFileWriter file;
	if(FAILED(file.Initialize(logPath)))
		return;
	file.Seek(0, FILE_END);
	file.WriteFormatted(
		"\"%02d:%02d:%02d\",\"%s\",%d,%d,\"%s\",%d,%d\x13\x10",
		locTime.wHour, locTime.wMinute, locTime.wSecond,
		lprof, lid, lwin, rprof, rid, rwin);
}

static void Swr_OnKO()
{
	int ret;
	char lprof[0x21], rprof[0x21];
	char *plprof = lprof, *prprof = rprof;
	if((ret = SwrGetParam(SWRPARAM_LPROFNAME)) == -1) lprof[0] = 0;
	else ::lstrcpyn(lprof, (LPCSTR)ret, 0x20); lprof[0x20] = 0;
	if((ret = SwrGetParam(SWRPARAM_RPROFNAME)) == -1) rprof[0] = 0;
	else ::lstrcpyn(rprof, (LPCSTR)ret, 0x20); rprof[0x20] = 0;

	int lwin = SwrGetParam(SWRPARAM_LWINCOUNT);
	int rwin = SwrGetParam(SWRPARAM_RWINCOUNT);
	int lid  = SwrGetParam(SWRPARAM_LCHARID);
	int rid  = SwrGetParam(SWRPARAM_RCHARID);
	int comm = SwrGetParam(SWRPARAM_COMMMODE);
	/* 臒l`FbN */
	if(lwin < 0 || lwin > 2 || rwin < 0 || rwin > 2) return;
	if(lid < 0 || lid >= SWRCHAR_MAX || rid < 0 || rid >= SWRCHAR_MAX) return;

	if(comm == SWRCOMMMODE_WATCH) return;
	if(comm == SWRCOMMMODE_CLIENT) {
		/* ɂ */
		int temp; char *stemp;
		temp = lwin ; lwin = rwin; rwin = temp;
		temp = lid; lid = rid; rid = temp;
		stemp = plprof; plprof = rprof; prprof = stemp;
	}

	/* A탊~b^ */
	if(s_tcgLimit && ++s_tcgCount > 3)
		return;

	/* ΐ\XV */
	ScoreLine_Append(lid, rid, rwin == 2);
	::PostMessage(s_hScoreLineDlg, UM_UPDATESCORELINE, 0, 0);
	/* O */
	if(s_isHage) Swr_OutputLog(lprof, lid, lwin, rprof, rid, rwin);


}

static void MainWindow_OnSwrCallback(HWND hwnd, WORD Msg, WORD param1, LPARAM param2)
{
	if(s_disableSwrCallback) return;
	if(
	Msg == SWRMSG_PARAMCHANGE &&		/* p[^ω */
	param1 == SWRPARAM_TOADDR)			/* ΐIPω */
		s_tcgCount = 0;	/* ~b^JEgZbg */

	if(
	Msg == SWRMSG_PARAMCHANGE &&		/* p[^ω */
	param1 == SWRPARAM_BATTLEMODE &&    /* og[hω */
	param2 == 4) {						/* og[hKO */
		int scene = SwrGetParam(SWRPARAM_SCENE);
		/* V[͒ʐMΐ풆 */
		if(scene == SWRSCENE_BATTLESV || scene == SWRSCENE_BATTLECL) {
			Swr_OnKO();
		}
	}
}

static void NotifyMenu_OnExit(HWND hwnd)
{
	::DestroyWindow(hwnd);
}

static void NotifyMenu_OnShowScoreLine(HWND hwnd)
{
	::ShowWindow(s_hScoreLineDlg, SW_SHOW);
	::SetForegroundWindow(s_hScoreLineDlg);
}

static void NotifyMenu_OnDisableSwrCallback(HWND hwnd)
{
	s_disableSwrCallback = !s_disableSwrCallback;
	::CheckMenuItem(s_hPopupMenu, ID_DISABLE_SWRCB,
		s_disableSwrCallback ? MF_CHECKED : MF_UNCHECKED);
}

static void NotifyMenu_OnTcgLimitter(HWND hwnd)
{
	s_tcgLimit = !s_tcgLimit;
	s_tcgCount = 0;
	::CheckMenuItem(s_hPopupMenu, ID_TCG_LIMITTER,
		s_tcgLimit ? MF_CHECKED : MF_UNCHECKED);
}

static BOOL MainWindow_ChangeMode(int mode)
{
	switch(s_mode = mode) {
	case 0:
		s_hScoreLineDlg = CreateDialogA(
			g_hInstance,
			MAKEINTRESOURCE(IDD_SCORELINE),
			NULL, ScoreLineDialog_DlgProc);
		return s_hScoreLineDlg != NULL;
	case 1:
		s_hScoreLineDlg = CreateDialogA(
			g_hInstance,
			MAKEINTRESOURCE(IDD_SCORELINE_QSR),
			NULL, ScoreLineDialog_DlgProc);
		return s_hScoreLineDlg != NULL;
	case 2:
		s_hScoreLineDlg = CreateDialogA(
			g_hInstance,
			MAKEINTRESOURCE(IDD_SCORELINE_QIB),
			NULL, ScoreLineQIBDialog_DlgProc);
		return s_hScoreLineDlg != NULL;
	}
	return FALSE;
}

static BOOL MainWindow_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	char profPathBuff[1025];
	g_settings.ReadString("General", "Profile", "", profPathBuff, sizeof profPathBuff);

	bool createProfile;
	Minimal::ProcessHeapPathA profPath;
	if(profPathBuff[0] == 0) {
		profPath = g_appPath;
		profPath /= "Default.hit";
		createProfile = true;
	} else {
		profPath = profPathBuff;
		createProfile = false;
	}
	if(!ScoreLine_Map(profPath, createProfile)) {
		OPENFILENAME ofn;
		ZeroMemory(&ofn, sizeof ofn);
		ofn.lStructSize = sizeof ofn;
		ofn.hwndOwner = hwnd;
		ofn.lpstrFile = profPathBuff;
		ofn.nMaxFile = sizeof profPathBuff;
		ofn.lpstrDefExt = "hit";
		ofn.lpstrFilter = "Hikouseki Trackrecord (*.hit)\0*.hit\0";
		ofn.Flags = OFN_CREATEPROMPT | OFN_NOCHANGEDIR;
		do {
			::MessageBox(hwnd, "vt@C̃}bsOɎs܂B", NULL, MB_OK | MB_ICONSTOP);
			if(!::GetOpenFileName(&ofn)) return FALSE;
		} while(!ScoreLine_Map(profPathBuff));
	}

	s_WM_TASKBAR_CREATED = RegisterWindowMessage("TaskbarCreated");

	s_nid.cbSize = sizeof(s_nid);
	s_nid.uID = 0;
	s_nid.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	s_nid.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	s_nid.hWnd = hwnd;
	s_nid.uCallbackMessage = UM_NOTIFYICON;
	lstrcpy(s_nid.szTip, WINDOW_TEXT);
	if(!Shell_NotifyIcon(NIM_ADD, &s_nid)) return FALSE;

	HMENU hDummyMenu = LoadMenu(g_hInstance, MAKEINTRESOURCE(IDR_MENU));
	if(hDummyMenu == NULL) return FALSE;
	s_hPopupMenu = GetSubMenu(hDummyMenu, 0);
	if(s_hPopupMenu == NULL) return FALSE;

	int mode = g_settings.ReadInteger("General", "Mode", 0);
	if(mode >= 0 && mode <= ID_MODE_QIB)
		::CheckMenuItem(s_hPopupMenu, ID_MODE_STANDARD + mode, MF_CHECKED);
	if(!MainWindow_ChangeMode(mode)) return FALSE;

	if(!SwrAddrInit(hwnd, UM_SWRCALLBACK)) return FALSE;

	s_tcgLimit = g_settings.ReadInteger("General", "tcgLimit", 0) != 0;
	s_tcgCount = 0;
	::CheckMenuItem(s_hPopupMenu, ID_TCG_LIMITTER,
		s_tcgLimit ? MF_CHECKED : MF_UNCHECKED);

	char Shacho[8];
	g_settings.ReadString("Gxxxxo", "Shacho", "", Shacho, sizeof Shacho);
	s_isHage = (::lstrcmpi(Shacho, "Hage") == 0);

	return TRUE;
}

static void MainWindow_OnDestroy(HWND hwnd)
{
	if(s_hScoreLineDlg)
		DestroyWindow(s_hScoreLineDlg);

	g_settings.WriteString("General", "Profile", ScoreLine_GetPath());
	g_settings.WriteInteger("General", "tcgLimit", s_tcgLimit);
	g_settings.WriteInteger("General", "Mode", s_mode);
	ScoreLine_Unmap();
	if(s_nid.cbSize) {
		Shell_NotifyIcon(NIM_DELETE, &s_nid);
	}
	SwrAddrFinish();

	::PostQuitMessage(0);
}

static void NotifyMenu_OnMode(HWND hwnd, int nID)
{
	CheckMenuItem(s_hPopupMenu, nID, MF_CHECKED);
	for(int i = ID_MODE_STANDARD; i <= ID_MODE_QIB; ++i)
		if(i != nID) ::CheckMenuItem(s_hPopupMenu, i, MF_UNCHECKED);

	int mode = nID - ID_MODE_STANDARD;
	if(s_mode != mode) {
		DestroyWindow(s_hScoreLineDlg);
		if(!MainWindow_ChangeMode(mode))
			::DestroyWindow(hwnd);
	}
}

static void MainWindow_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch(id) {
	case ID_MODE_STANDARD:	
	case ID_MODE_QSR:
	case ID_MODE_QIB:		NotifyMenu_OnMode(hwnd, id); break;
	case ID_TCG_LIMITTER:	NotifyMenu_OnTcgLimitter(hwnd); break;
	case ID_DISABLE_SWRCB:	NotifyMenu_OnDisableSwrCallback(hwnd); break;
	case ID_SHOW_SCORELINE: NotifyMenu_OnShowScoreLine(hwnd); break;
	case ID_EXIT:			NotifyMenu_OnExit(hwnd); break;
	}
}

static BOOL MainWindow_OnNotifyIcon(HWND hwnd, WPARAM id, LPARAM Msg)
{
	if(Msg == WM_RBUTTONDOWN) {
		POINT curPos;
		::GetCursorPos(&curPos);
		::SetForegroundWindow(hwnd);
		::TrackPopupMenu(s_hPopupMenu, TPM_LEFTBUTTON | TPM_TOPALIGN, curPos.x, curPos.y, 0, hwnd, NULL);
		::SendMessage(hwnd, WM_NULL, 0, 0);
	} else if(Msg == WM_LBUTTONDBLCLK) {
		NotifyMenu_OnShowScoreLine(hwnd);
	}
	return TRUE;
}

static BOOL MainWindow_OnUnhandled(HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	if(Msg == s_WM_TASKBAR_CREATED) {
		::Shell_NotifyIcon(NIM_ADD, &s_nid);
		return TRUE;
	}
	else
	return FALSE;
}

LRESULT CALLBACK MainWindow_WndProc(HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch(Msg) {
	HANDLE_MSG(hwnd, WM_CREATE,   MainWindow_OnCreate);
	HANDLE_MSG(hwnd, WM_DESTROY,  MainWindow_OnDestroy);
	HANDLE_MSG(hwnd, WM_COMMAND,  MainWindow_OnCommand);
	case UM_NOTIFYICON:  return MainWindow_OnNotifyIcon(hwnd, wParam, lParam);
	case UM_SWRCALLBACK: return MainWindow_OnSwrCallback(hwnd, LOWORD(wParam), HIWORD(wParam), lParam), 0L;
	}
	return MainWindow_OnUnhandled(hwnd, Msg, wParam, lParam) ? 0
		 : ::DefWindowProc(hwnd, Msg, wParam, lParam);

}

bool MainWindow_Initialize()
{
	WNDCLASS wc;
	ZeroMemory(&wc, sizeof wc);
	wc.hInstance = g_hInstance;
	wc.lpfnWndProc = MainWindow_WndProc;
	wc.lpszClassName = WINDOW_CLASS;
	if(::RegisterClass(&wc) == (ATOM)0) return false;

	HWND hMainWnd = ::CreateWindowA(
		WINDOW_CLASS, NULL, WS_POPUP,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, g_hInstance, NULL);
	if(!hMainWnd) return false;

	return true;
}
